﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.EGO {
  internal static class EgoUtilities {
    public static int ArgMax<T>(this IEnumerable<T> values, Func<T, double> func) {
      var max = double.MinValue;
      var maxIdx = 0;
      var idx = 0;
      foreach (var v in values) {
        var d = func.Invoke(v);
        if (d > max) {
          max = d;
          maxIdx = idx;
        }
        idx++;
      }
      return maxIdx;
    }
    public static int ArgMin<T>(this IEnumerable<T> values, Func<T, double> func) {
      return ArgMax(values, x => -func.Invoke(x));
    }

    public static ResultCollection SyncRunSubAlgorithm(IAlgorithm alg, int random) {
      EventWaitHandle trigger = new AutoResetEvent(false);
      Exception ex = null;
      EventHandler<EventArgs<Exception>> exhandler = (sender, e) => ex = e.Value;
      EventHandler stoppedHandler = (sender, e) => trigger.Set();
      alg.ExceptionOccurred += exhandler;
      alg.Stopped += stoppedHandler;
      alg.Prepare();
      alg.Start();
      trigger.WaitOne();
      alg.ExceptionOccurred -= exhandler;
      alg.Stopped -= stoppedHandler;
      if (ex != null) throw ex;
      return alg.Results;
    }
    public static RealVector[] GetUniformRandomDesign(int points, int dim, DoubleMatrix bounds, IRandom random) {
      var res = new RealVector[points];
      for (var i = 0; i < points; i++) {
        var r = new RealVector(dim);
        res[i] = r;
        for (var j = 0; j < dim; j++) {
          var b = j % bounds.Rows;
          r[j] = UniformRandom(bounds[b, 0], bounds[b, 1], random);
        }
      }
      return res;
    }
    public static double UniformRandom(double min, double max, IRandom rand) {
      return rand.NextDouble() * (max - min) + min;
    }

    public static double GetEstimation(this IRegressionModel model, RealVector r) {
      var dataset = GetDataSet(new[] { new Tuple<RealVector, double>(r, 0.0) });
      return model.GetEstimatedValues(dataset, new[] { 0 }).First();
    }
    public static double GetVariance(this IConfidenceRegressionModel model, RealVector r) {
      var dataset = GetDataSet(new[] { new Tuple<RealVector, double>(r, 0.0) });
      return model.GetEstimatedVariances(dataset, new[] { 0 }).First();
    }

    public static Dataset GetDataSet(IReadOnlyList<Tuple<RealVector, double>> samples) {
      var n = samples[0].Item1.Length + 1;
      var data = new double[samples.Count, n];
      var names = new string[n - 1];
      for (var i = 0; i < n; i++)
        if (i < names.Length) {
          names[i] = "input" + i;
          for (var j = 0; j < samples.Count; j++) data[j, i] = samples[j].Item1[i];
        } else
          for (var j = 0; j < samples.Count; j++) data[j, n - 1] = samples[j].Item2;
      return new Dataset(names.Concat(new[] { "output" }).ToArray(), data);
    }
  }
}
